//**************************************************************************
//*                     This file is part of the                           *
//*                      Mpxplay - audio player.                           *
//*                  The source code of Mpxplay is                         *
//*        (C) copyright 1998-2005 by PDSoft (Attila Padar)                *
//*                    http://mpxplay.cjb.net                              *
//*                  email: mpxplay@freemail.hu                            *
//**************************************************************************
//*  This program is distributed in the hope that it will be useful,       *
//*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
//*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                  *
//*  Please contact with the author (with me) if you want to use           *
//*  or modify this source.                                                *
//**************************************************************************
//function: WavPack file handling in DLL
//requires the libwavpa\libwavpa.lib and wavpack.h files

#include <in_file.h>
#include <in_funcs.h>
#include <newfunc\newfunc.h>
#include <libwavp\wavpack.h>

#ifdef WAVPACK_DEBUG
 #include <stdarg.h>
#endif

typedef struct wavpack_decoder_data{
 WavpackContext *wpc;
 unsigned int  bytes_per_sample;
}wavpack_decoder_data;

struct mpxplay_resource_s *mrs;

static void wavpack_infile_close(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,void *infile_data,struct mpxplay_infile_info_s *miis);

static int wavpack_assign_values(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,struct wavpack_decoder_data *wpdi,struct mpxplay_infile_info_s *miis)
{
 struct mpxplay_audio_decoder_info_s *adi=miis->audio_decoder_infos;
 unsigned int encmode;

 adi->filechannels = adi->outchannels = WavpackGetReducedChannels(wpdi->wpc);//WavpackGetNumChannels(wpdi->wpc);
 if((adi->outchannels<PCM_MIN_CHANNELS) || (adi->outchannels>PCM_MAX_CHANNELS))
  return 0;

 adi->bits = WavpackGetBitsPerSample(wpdi->wpc);
 if((adi->bits<PCM_MIN_BITS) || (adi->bits>PCM_MAX_BITS))
  return 0;

 adi->freq = WavpackGetSampleRate(wpdi->wpc);
 wpdi->bytes_per_sample = WavpackGetBytesPerSample(wpdi->wpc);
 adi->pcmdatalen=WavpackGetNumSamples(wpdi->wpc);

 encmode=WavpackGetMode(wpdi->wpc);
 if(encmode&MODE_FLOAT){
  adi->infobits|=ADI_FLAG_FLOATOUT;
  adi->bits=1;
  wpdi->bytes_per_sample=sizeof(MPXPLAY_PCMOUT_FLOAT_T);
 }

 adi->longname="WavPack";

 if(encmode&MODE_HYBRID){
  adi->bitrate=(long)((float)miis->filesize*8.0/1000.0*(float)adi->freq/(float)adi->pcmdatalen);
 }else{
  long compr_ratio;
  adi->bitratetext=malloc(MPXPLAY_ADITEXTSIZE_BITRATE+8);
  if(!adi->bitratetext)
   return 0;
  compr_ratio=(long)(1000.0*(float)miis->filesize/(float)adi->pcmdatalen/(float)wpdi->bytes_per_sample/(float)adi->filechannels);
  sprintf(adi->bitratetext,"%2d/%2.2d.%1.1d%%",adi->bits,compr_ratio/10,compr_ratio%10);
 }

 return 1;
}

static void *wavpack_infile_check(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,char *filename,struct mpxplay_infile_info_s *miis)
{
 struct wavpack_decoder_data *wpdi=NULL;
 char errstr[128];

 if(!fbfs->fopen_read(fbds,filename,0))
  goto err_out_check;

 miis->filesize=fbfs->filelength(fbds);
 if(miis->filesize<16)  // ???
  goto err_out_check;

 wpdi=(struct wavpack_decoder_data *)calloc(1,sizeof(struct wavpack_decoder_data));
 if(!wpdi)
  goto err_out_check;

 wpdi->wpc=(WavpackContext *)calloc(1,sizeof(WavpackContext));
 if(!wpdi->wpc)
  goto err_out_check;

 if(!WavpackOpenFileInput(wpdi->wpc, fbfs->fread, fbds, errstr))
  goto err_out_check;

 if(!wavpack_assign_values(fbfs,fbds,wpdi,miis))
  goto err_out_check;

 return wpdi;

err_out_check:
 wavpack_infile_close(fbfs,fbds,wpdi,miis);
 return NULL;
}

static void wavpack_infile_close(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,void *infile_data,struct mpxplay_infile_info_s *miis)
{
 struct mpxplay_audio_decoder_info_s *adi=miis->audio_decoder_infos;
 struct wavpack_decoder_data *wpdi=infile_data;
 if(wpdi){
  if(wpdi->wpc)
   free(wpdi->wpc);
  free(wpdi);
 }
 if(adi->bitratetext)
  free(adi->bitratetext);
 fbfs->fclose(fbds);
}

//-------------------------------------------------------------------------

static void format_samples(int bps, unsigned char *dst, long *src, unsigned long samcnt)
{
 if(bps==1){
  do{
   *dst++ = *src++ + 128;
  }while(--samcnt);
 }else{
  do{
   *((long *)dst)=*src++;
   dst+=bps;
  }while(--samcnt);
 }
}

static int wavpack_infile_decode(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,void *infile_data,struct mpxplay_infile_info_s *miis)
{
 mpxplay_audio_decoder_info_s *adi=miis->audio_decoder_infos;
 struct wavpack_decoder_data *wpdi=infile_data;

 WavpackUpdateStreamhand(wpdi->wpc,fbds);

 adi->pcmoutsamplenum=WavpackUnpackSamples(wpdi->wpc,(long *)adi->pcm_outdatabuf,adi->pcmsamplenum_frame);
 if(!adi->pcmoutsamplenum)
  return MPXPLAY_ERROR_INFILE_NODATA;
 adi->pcmoutsamplenum*=adi->outchannels;

 if(wpdi->bytes_per_sample<4)
  format_samples(wpdi->bytes_per_sample,(unsigned char *)adi->pcm_outdatabuf,(long *)adi->pcm_outdatabuf,adi->pcmoutsamplenum);

 return MPXPLAY_ERROR_INFILE_OK;
}

//-------------------------------------------------------------------------

static void wavpack_clearbuffs(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,void *infile_data,struct mpxplay_infile_info_s *miis,int cleartype)
{
 struct wavpack_decoder_data *wpdi=infile_data;
 WavpackResetDecoding(wpdi->wpc);
}

static long wavpack_fseek(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,void *infile_data,struct mpxplay_infile_info_s *miis,long newmpxframenum)
{
 struct wavpack_decoder_data *wpdi=infile_data;
 long newfilepos;

 WavpackUpdateStreamhand(wpdi->wpc,fbds);

 newfilepos=(long)((float)newmpxframenum*(float)miis->filesize/(float)miis->allframes);
 if(fbfs->fseek(fbds,newfilepos,SEEK_SET)<0)
  return MPXPLAY_ERROR_INFILE_EOF;
 return newmpxframenum;
}

/*void wavpack_debugf(const char *format, ...)
{
#ifdef WAVPACK_DEBUG
 va_list ap;
 char sout[500];

 va_start(ap,format);
 vsprintf(sout, format, ap );
 va_end(ap);

 mrs->pds_textdisplay_printf(sout);
#endif
}*/

struct mpxplay_infile_func_s IN_WAVPACK_funcs={
 0,
 NULL,
 NULL,
 &wavpack_infile_check,
 &wavpack_infile_check,
 &wavpack_infile_check,
 &wavpack_infile_close,
 &wavpack_infile_decode,
 &wavpack_fseek,
 &wavpack_clearbuffs,
 &ape_tag_get,
 &ape_tag_put,
 NULL,
 {"WV",NULL}
};

static mpxplay_module_entry_s faad_module_entry={
 MPXPLAY_DLLMODULETYPE_FILEIN_PARSER,
 0,
 "WavPack",
 MPXPLAY_DLLMODULEVER_FILEIN_PARSER,
 (void *)&IN_WAVPACK_funcs
};

static mpxplay_dll_entry_s mpxplay_dll_entry_structure={
 MPXPLAY_DLLENTRY_STRUCTURE_VERSION,
 {
  &faad_module_entry,
  NULL
 }
};

extern void dllstrtr_update_crwdata(unsigned long *cwd);

long __export mpxplay_dll_entrypoint(struct mpxplay_resource_s *p_mrs,unsigned long *crwdata_begin)
{
 mrs=p_mrs;
 dllstrtr_update_crwdata(crwdata_begin);
 return (long)(&mpxplay_dll_entry_structure);
}
