//**************************************************************************
//*                     This file is part of the                           *
//*                      Mpxplay - audio player.                           *
//*                  The source code of Mpxplay is                         *
//*        (C) copyright 1998-2005 by PDSoft (Attila Padar)                *
//*                    http://mpxplay.cjb.net                              *
//*                  email: mpxplay@freemail.hu                            *
//**************************************************************************
//*  This program is distributed in the hope that it will be useful,       *
//*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
//*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                  *
//*  Please contact with the author (with me) if you want to use           *
//*  or modify this source.                                                *
//**************************************************************************
//function: tag read/write

#include <newfunc\newfunc.h>
#include <in_funcs.h>
#include <dec_ape\apeinfo.h>

extern struct mpxplay_resource_s *mrs;

//--------------------------------------------------------------------------
//ID3v1

static char *id3v1genres[]={
  "Blues", "Classic Rock", "Country", "Dance", "Disco", "Funk", "Grunge",
  "Hip-Hop", "Jazz", "Metal", "New Age", "Oldies", "Other", "Pop", "R&B",
  "Rap", "Reggae", "Rock", "Techno", "Industrial", "Alternative", "Ska",
  "Death Metal", "Pranks", "Soundtrack", "Euro-Techno", "Ambient",
  "Trip-Hop", "Vocal", "Jazz+Funk", "Fusion", "Trance", "Classical",
  "Instrumental", "Acid", "House", "Game", "Sound Clip", "Gospel", "Noise",
  "AlternRock", "Bass", "Soul", "Punk", "Space", "Meditative",
  "Instrumental Pop", "Instrumental Rock", "Ethnic", "Gothic", "Darkwave",
  "Techno-Industrial", "Electronic", "Pop-Folk", "Eurodance", "Dream",
  "Southern Rock", "Comedy", "Cult", "Gangsta", "Top 40", "Christian Rap",
  "Pop/Funk", "Jungle", "Native American", "Cabaret", "New Wave",
  "Psychadelic", "Rave", "Showtunes", "Trailer", "Lo-Fi", "Tribal",
  "Acid Punk", "Acid Jazz", "Polka", "Retro", "Musical", "Rock & Roll",
  "Hard Rock", "Folk", "Folk/Rock", "National Folk", "Swing", "Fast-Fusion",
  "Bebob", "Latin", "Revival", "Celtic", "Bluegrass", "Avantgarde",
  "Gothic Rock", "Progressive Rock", "Psychedelic Rock", "Symphonic Rock",
  "Slow Rock", "Big Band", "Chorus", "Easy Listening", "Acoustic", "Humour",
  "Speech", "Chanson", "Opera", "Chamber Music", "Sonata", "Symphony",
  "Booty Bass", "Primus", "Porn Groove", "Satire", "Slow Jam", "Club",
  "Tango", "Samba", "Folklore", "Ballad", "Power Ballad", "Rhythmic Soul",
  "Freestyle", "Duet", "Punk Rock", "Drum Solo", "A capella", "Euro-House",
  "Dance Hall", "Goa", "Drum & Bass", "Club House", "Hardcore", "Terror",
  "Indie", "BritPop", "NegerPunk", "Polsk Punk", "Beat", "Christian Gangsta",
  "Heavy Metal", "Black Metal", "Crossover", "Contemporary C",
  "Christian Rock", "Merengue", "Salsa", "Thrash Metal", "Anime", "JPop",
  "SynthPop"
};

#define MAX_ID3GENRENUM (sizeof(id3v1genres)/sizeof(char *))

static unsigned int id3v10_partlens[5]={30,30,30,4,30};

unsigned int mpx_tag_check_id3v1(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds)
{
 char tag[4];
 if(fbfs->fseek(fbds,-128,SEEK_END)<0)
  return 0;
 if(fbfs->fread(fbds,tag,3)!=3)
  return 0;
 if(tag[0]=='T' && tag[1]=='A' && tag[2]=='G')
  return 1;
 return 0;
}

//always call check_id3tag_v1 before this
char *mpx_tag_read_id3v1(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,char **id3ip,char *id3p,struct mpxplay_textconv_func_s *mpxplay_textconv_funcs)
{
 unsigned int i,partlen,reallen;
 char *readp,readtmp[128];

 readp=&readtmp[0];
 if(fbfs->fread(fbds,readp,128-3)!=(128-3))
  return id3p;

 for(i=0;i<5;i++){
  partlen=id3v10_partlens[i];
  if(!id3ip[i]){ // maybe we've got it from ID3v2
   mrs->pds_strncpy(id3p,readp,partlen);
   if(i==I3I_COMMENT){
    if(id3p[partlen-1]<32) // tracknumber check
     id3p[partlen-1]=0;
   }
   id3p[partlen]=0;
   reallen=mrs->pds_strlenc(id3p,32);
   if(reallen){
    id3ip[i]=id3p;
    id3p[reallen]=0;
    reallen=mpxplay_textconv_funcs->all_to_char(id3p,reallen,0);
    id3p+=reallen+1;
   }
  }
  readp+=partlen;
 }
 //tracknumber from id3v1.1
 if(!id3ip[I3I_TRACKNUM]){  // maybe we've got it from ID3v2
  i=(unsigned int)readp[-1];
  if(i && !readp[-2]){
   id3ip[I3I_TRACKNUM]=id3p;
   id3p+=sprintf(id3p,"%d",i)+1;
  }
 }
 //genre
 if(!id3ip[I3I_GENRE]){     // maybe we've got it from ID3v2
  i=(unsigned int)readp[0];
  if(i<MAX_ID3GENRENUM)
   id3ip[I3I_GENRE]=id3v1genres[i];
 }
 return id3p;
}

static unsigned int tag_id3v1_get_genrenum(char *genrename)
{
 unsigned int i;

 if(genrename)
  for(i=0;i<MAX_ID3GENRENUM;i++)
   if(stricmp(id3v1genres[i],genrename)==0)
    return i;

 return 255;
}

int mpx_tag_write_id3v1(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,char **id3ip)
{
 unsigned int error=MPXPLAY_ERROR_INFILE_WRITETAG_UNKNOWN;
 long fileoffset;
 char strtemp[192];

 if(mpx_tag_check_id3v1(fbfs,fbds))        // is file tagged already?
  fileoffset=fbfs->fseek(fbds,-128,SEEK_END);
 else
  fileoffset=fbfs->fseek(fbds,   0,SEEK_END);
 if(fileoffset>0){                // successfull file positioning
  if(id3ip[I3I_TRACKNUM] && id3ip[I3I_TRACKNUM][0]){       // ID3v1.1
   unsigned int tracknum=mrs->pds_atol(id3ip[I3I_TRACKNUM]);
   if(tracknum>255)
    tracknum=255;
   sprintf(strtemp,"TAG%-30.30s%-30.30s%-30.30s%-4.4s%-28.28s%c%c%c",
                   (id3ip[I3I_TITLE])?  id3ip[I3I_TITLE]:"",
                   (id3ip[I3I_ARTIST])? id3ip[I3I_ARTIST]:"",
                   (id3ip[I3I_ALBUM])?  id3ip[I3I_ALBUM]:"",
                   (id3ip[I3I_YEAR])?   id3ip[I3I_YEAR]:"",
                   (id3ip[I3I_COMMENT])?id3ip[I3I_COMMENT]:"",0,
                   tracknum,
                   tag_id3v1_get_genrenum(id3ip[I3I_GENRE]));
  }else{                                                   // ID3v1.0
   sprintf(strtemp,"TAG%-30.30s%-30.30s%-30.30s%-4.4s%-30.30s%c",
                   (id3ip[I3I_TITLE])?  id3ip[I3I_TITLE]:"",
                   (id3ip[I3I_ARTIST])? id3ip[I3I_ARTIST]:"",
                   (id3ip[I3I_ALBUM])?  id3ip[I3I_ALBUM]:"",
                   (id3ip[I3I_YEAR])?   id3ip[I3I_YEAR]:"",
                   (id3ip[I3I_COMMENT])?id3ip[I3I_COMMENT]:"",
                   tag_id3v1_get_genrenum(id3ip[I3I_GENRE]));
  }
  if(fbfs->fwrite(fbds,strtemp,128)==128) // successfull TAG writing
   error=MPXPLAY_ERROR_INFILE_OK;
 }
 return error;
}

//---------------------------------------------------------------------
//APETag

typedef struct tagtype_s{
 char *name;
 unsigned int i3i_index;
}tagtype_s;

static tagtype_s tagtypes[]={
 {"Artist",I3I_ARTIST},{"Title"  ,I3I_TITLE}  ,{"Album",I3I_ALBUM},
 {"Year"  ,I3I_YEAR}  ,{"Comment",I3I_COMMENT},{"Genre",I3I_GENRE},
 {"Track" ,I3I_TRACKNUM}
};

#define TAGTYPENUM (sizeof(tagtypes)/sizeof(struct tagtype_s))

static int ape_tag_check_apetag(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,struct APETag_s *apetag)
{
 int version,taglen;
 long filesize=fbfs->filelength(fbds);
 if(filesize<=((long)sizeof(struct APETag_s)))
  return 0;
 if(fbfs->fseek(fbds,-((long)sizeof(struct APETag_s)),SEEK_END)<0)
  return 0;
 if(fbfs->fread(fbds,(char *)apetag,sizeof(struct APETag_s))!=sizeof(struct APETag_s))
  return 0;
 if(mrs->pds_strncmp(apetag->ID,"APETAGEX",sizeof(apetag->ID))!=0)
  return 0;
 version=read_le_long(apetag->Version);
 if(version!=1000 && version!=2000)
  return 0;
 taglen=read_le_long(apetag->Length);
 if(taglen<=sizeof(struct APETag_s) || (taglen>=filesize))
  return 0;
 if(!read_le_long(apetag->TagCount))
  return 0;
 if(fbfs->fseek(fbds,-taglen,SEEK_END)<0)
  return 0;
 return 1;
}

static char *ape_tag_read_apetag(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,struct APETag_s *apetag,char **id3ip,char *id3p,struct mpxplay_textconv_func_s *mpxplay_textconv_funcs)
{
 long tt,version,taglen,tagcount;
 long namelen,datalen,dataflags;
 char *tagbuff,*tbp,*tname,*tdata;

 taglen=read_le_long(apetag->Length);
 tagbuff=malloc(taglen);
 if(!tagbuff)
  return id3p;
 if(fbfs->fread(fbds,tagbuff,taglen-sizeof(struct APETag_s))!=(taglen-sizeof(struct APETag_s)))
  goto err_out_tagread;
 tbp=tagbuff;
 tagbuff[taglen-sizeof(struct APETag_s)]=0;
 version=read_le_long(apetag->Version);
 tagcount=read_le_long(apetag->TagCount);
 for(;*tbp && tagcount--;){
  datalen  =read_le_long(tbp); tbp+=4;
  dataflags=read_le_long(tbp); tbp+=4;
  namelen=strlen(tbp)+1;
  if(namelen>128)        // impossible data
   break;
  tname=tbp; tbp+=namelen;
  tdata=tbp; tbp+=datalen;
  if(tbp>(&tagbuff[taglen-sizeof(struct APETag_s)])) // bad data -> overflow
   break;
  if((namelen>1) && ((version==1000 && datalen>1) || (version==2000 && datalen>0 && !(dataflags&2)))){
   for(tt=0;tt<TAGTYPENUM;tt++){
    int i3iindex=tagtypes[tt].i3i_index;
    if(!id3ip[i3iindex] && mrs->pds_stricmp(tname,tagtypes[tt].name)==0){
     id3ip[i3iindex]=id3p;
     mrs->pds_memcpy(id3p,tdata,datalen);
     if(version==1000){
      id3p+=mpxplay_textconv_funcs->all_to_char(id3p,datalen-1,0)+1; // datalen includes the terminating null in v1
     }else{
      id3p[datalen]=0;
      if((*(mpxplay_textconv_funcs->control))&ID3TEXTCONV_UTF_AUTO)
       datalen=mpxplay_textconv_funcs->utf8_to_char(id3p,datalen);  // allways in v2
      id3p+=mpxplay_textconv_funcs->all_to_char(id3p,datalen,ID3TEXTCONV_UTF8)+1;   // datalen does not include the terminating null in v2
     }
     break;
    }
   }
  }
 }
err_out_tagread:
 free(tagbuff);
 return id3p;
}

char *ape_tag_get(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,void *infile_data,char **id3ip,char *id3p,struct mpxplay_textconv_func_s *mpxplay_textconv_funcs)
{
 struct APETag_s apetag;

 if(mpx_tag_check_id3v1(fbfs,fbds))
  return mpx_tag_read_id3v1(fbfs,fbds,id3ip,id3p,mpxplay_textconv_funcs);

 if(ape_tag_check_apetag(fbfs,fbds,&apetag))
  return ape_tag_read_apetag(fbfs,fbds,&apetag,id3ip,id3p,mpxplay_textconv_funcs);

 return id3p;
}

int ape_tag_put(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,char **id3ip)
{
 int tt,i,extratags=0,cutfile=0;
 long alltagsize=0,tagcount=0,newalltagsize,*tagdlens=NULL,*tagflags=NULL;
 __int32 tagnamelen,tagdatalen,flags=0,version;
 char *tagbuff=NULL,*tbp,**tagnames=NULL,**tagdatas=NULL;
 struct APETag_s apetag;

 if(mpx_tag_check_id3v1(fbfs,fbds))
  return mpx_tag_write_id3v1(fbfs,fbds,id3ip);

 if(ape_tag_check_apetag(fbfs,fbds,&apetag)){ // APE file already has APETAGEX
  // load all tagdatas, update and write back
  version=read_le_long(apetag.Version);
  if(version!=1000)// && version!=2000)
   return MPXPLAY_ERROR_INFILE_WRITETAG_TAGTYPE;
  alltagsize=read_le_long(apetag.Length);
  tbp=tagbuff=malloc(alltagsize);
  tagcount=read_le_long(apetag.TagCount);
  tagnames=(char **)malloc((tagcount+TAGTYPENUM)*sizeof(char *));
  tagdatas=(char **)malloc((tagcount+TAGTYPENUM)*sizeof(char *));
  tagdlens=(long *)malloc((tagcount+TAGTYPENUM)*sizeof(long));
  tagflags=(long *)malloc((tagcount+TAGTYPENUM)*sizeof(long));
  if(!tagbuff || !tagnames || !tagdatas || !tagdlens || !tagflags)
   goto err_out_tagput;
  //load
  if(fbfs->fread(fbds,tagbuff,alltagsize-sizeof(struct APETag_s))!=(alltagsize-sizeof(struct APETag_s)))
   goto err_out_tagput;
  for(i=0;i<tagcount && (tbp<(tagbuff+alltagsize-sizeof(struct APETag_s)));i++){
   tagdlens[i]=read_le_long(tbp);tbp+=4;
   tagflags[i]=read_le_long(tbp);tbp+=4;
   tagnames[i]=tbp;              tbp+=strlen(tbp)+1;
   tagdatas[i]=tbp;              tbp+=tagdlens[i];
  }
  tagcount=i; // correction in the case of a wrong ape-tag
  //update
  for(tt=0;tt<TAGTYPENUM;tt++){
   int i3iindex=tagtypes[tt].i3i_index;
   if(id3ip[i3iindex]){
    for(i=0;i<tagcount;i++){
     if((version==1000 && tagdlens[i]>1) || (version==2000 && tagdlens[i]>0 && !(tagflags[i]&2))){
      if(mrs->pds_stricmp(tagnames[i],tagtypes[tt].name)==0){ // found, update
       tagdatas[i]=id3ip[i3iindex];
       tagdlens[i]=strlen(id3ip[i3iindex])+((version==1000)? 1:0);
       break;
      }
     }
    }
    if(i==tagcount){                                 // not found, new
     tagnames[tagcount+extratags]=tagtypes[tt].name;
     tagdatas[tagcount+extratags]=id3ip[i3iindex];
     tagdlens[tagcount+extratags]=strlen(id3ip[i3iindex])+((version==1000)? 1:0);
     tagflags[tagcount+extratags]=0;
     extratags++;
    }
   }
  }
  //write back
  if(fbfs->fseek(fbds,-alltagsize,SEEK_END)<0)
   goto err_out_tagput;
  newalltagsize=0;
  tagcount+=extratags;
  for(i=0;i<tagcount;i++){
   tagnamelen=strlen(tagnames[i])+1;
   fbfs->fwrite(fbds,&tagdlens[i],sizeof(tagdlens[i]));
   fbfs->fwrite(fbds,&tagflags[i],sizeof(tagflags[i]));
   fbfs->fwrite(fbds,tagnames[i],tagnamelen);
   fbfs->fwrite(fbds,tagdatas[i],tagdlens[i]);
   newalltagsize+=sizeof(tagdlens[i])+sizeof(tagflags[i])+tagnamelen+tagdlens[i];
  }
  if(newalltagsize<alltagsize)
   cutfile=1;
  alltagsize=newalltagsize;
  free(tagflags);
  free(tagdlens);
  free(tagdatas);
  free(tagnames);
  free(tagbuff);
 }else{
  // write an absolute new APEv1 tag
  version=1000;
  if(fbfs->fseek(fbds,0,SEEK_END)<0)
   return MPXPLAY_ERROR_INFILE_EOF;
  for(tt=0;tt<TAGTYPENUM;tt++){
   int i3iindex=tagtypes[tt].i3i_index;
   if(id3ip[i3iindex]){
    tagnamelen=strlen(tagtypes[tt].name)+1;
    tagdatalen=strlen(id3ip[i3iindex])+1;
    fbfs->fwrite(fbds,&tagdatalen,sizeof(tagdatalen));
    fbfs->fwrite(fbds,&flags,sizeof(flags));
    fbfs->fwrite(fbds,tagtypes[tt].name,tagnamelen);
    fbfs->fwrite(fbds,id3ip[i3iindex],tagdatalen);
    alltagsize+=sizeof(tagdatalen)+sizeof(flags)+tagnamelen+tagdatalen;
    tagcount++;
   }
  }
 }
 //finish/close the tag
 mrs->pds_memcpy(apetag.ID,"APETAGEX",sizeof("APETAGEX"));
 write_le_long(apetag.Version,version);
 write_le_long(apetag.Length,alltagsize+sizeof(struct APETag_s));
 write_le_long(apetag.TagCount,tagcount);
 write_le_long(apetag.Flags,0);
 memset(apetag.Reserved,0,sizeof(apetag.Reserved));
 fbfs->fwrite(fbds,(void *)&apetag,sizeof(struct APETag_s));

 if(cutfile)
  fbfs->chsize(fbds,fbfs->ftell(fbds));

 return MPXPLAY_ERROR_INFILE_OK;

err_out_tagput:
 if(tagflags)
  free(tagflags);
 if(tagdlens)
  free(tagdlens);
 if(tagdatas)
  free(tagdatas);
 if(tagnames)
  free(tagnames);
 if(tagbuff)
  free(tagbuff);
 return MPXPLAY_ERROR_INFILE_MEMORY;
}
